unit libraryu;

interface

uses
  Windows, Messages, SysUtils, Variants, Classes, Graphics, Controls, Forms,
  Dialogs, Grids, ExtCtrls, StdCtrls, TxtIntface, ComCtrls, Menus;

type
  TfmLibrary = class(TForm)
	dlgOpen: TOpenDialog;
    pnList: TPanel;
    Splitter1: TSplitter;
    pnComp: TPanel;
    gbProps: TGroupBox;
    pbComp: TPaintBox;
    sgNames: TStringGrid;
    pnFilter: TPanel;
    Label1: TLabel;
    edFilter: TEdit;
	tmRefresh: TTimer;
    Label2: TLabel;
    lPinCount: TLabel;
    Label3: TLabel;
    lUnits: TLabel;
    pnView: TPanel;
    Label5: TLabel;
	Label6: TLabel;
	Label7: TLabel;
    edAliases: TEdit;
	edKeywords: TEdit;
    edDescription: TEdit;
    mnGrid: TPopupMenu;
    mnDelete1: TMenuItem;
    mnCut1: TMenuItem;
    mnPaste1: TMenuItem;
	lCopy: TLabel;
    lMove: TLabel;
    dlgSave: TSaveDialog;
	pnZoom: TPanel;
	Label4: TLabel;
	lZoom: TLabel;
	tbZoom: TTrackBar;
	ckNames: TCheckBox;
	ckNumbers: TCheckBox;
	pnMorgan: TPanel;
	cbUnits: TComboBox;
	ckDeMorgan: TCheckBox;
    mnCopy1: TMenuItem;
    N1: TMenuItem;
    mnSortbyname1: TMenuItem;
    lFilename: TLabel;
	procedure FormCloseQuery(Sender: TObject; var CanClose: Boolean);
	procedure FormClose(Sender: TObject; var Action: TCloseAction);
	procedure LoadAskFile;
	procedure LoadFile(fn: string);
	constructor create(AOwner: TComponent); override;
	procedure refresh;
	procedure edFilterChange(Sender: TObject);
	procedure tmRefreshTimer(Sender: TObject);
	procedure sgNamesClick(Sender: TObject);
	procedure pbCompPaint(Sender: TObject);
	procedure tbZoomChange(Sender: TObject);
	procedure sgNamesDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
	procedure sgNamesDragDrop(Sender, Source: TObject; X, Y: Integer);
	procedure Save;
	procedure SaveAs;
	procedure mnDelete1Click(Sender: TObject);
	procedure DoLayout1;
	procedure DoLayout2;
    procedure mnCut1Click(Sender: TObject);
    procedure pnListResize(Sender: TObject);
	procedure mnPaste1Click(Sender: TObject);
	procedure mnCopy1Click(Sender: TObject);
	procedure mnSortbyname1Click(Sender: TObject);
	procedure edFilterKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
	procedure sgNamesKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
	procedure edKeywordsChange(Sender: TObject);
	procedure pbCompMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
	procedure pbCompMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
	procedure pbCompMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
	procedure SetModiCaption;
  private
	{ Private declarations }

  public
	{ Public declarations }
	libpath		: string;
	libname		: string;
	filter		: string;
	KiComps		: TList;
	ActComp		: TKiComponent;
	libmodified	: boolean;
	intmodifying: boolean;

	// viewing
	panx,pany	: integer;
	pandown		: boolean;
	opx,opy		: integer;
  end;

var
  fmLibrary: TfmLibrary;

function CmpPassFilter(cmp: TKiComponent; flt: string): boolean;

implementation
uses KiLibmanagerU;

{$R *.dfm}

procedure TfmLibrary.LoadFile(fn: string);
var
  f		: TextFile;
  st	: string;
  err	: integer;
  errsum: integer;
  comp	: TKiComponent;

  function getline: boolean;	// true if error
  begin
	result := true;
	if eof(f) then begin
	  err := 1;
	  exit
	end;
	readln(f, st);
	err := IOResult;
	if errsum=0 then errsum := err;
	if err=0 then result := false
  end;

label goterror;

begin
  errsum  := 0;
  assignfile(f, fn);
  reset(f);
  err := IOResult;
  if err<>0 then goto goterror;
  if getline then goto goterror;
  if not IsKiCadLibProlog(st) then application.MessageBox('Sure it is a library?','');

  libname := extractfilename(fn);
  libpath := extractfiledir(fn);
  
  while not eof(f) do begin
	getline;
	// if (st='') or (st[1]='#') then continue;

	if copy(st,1,4)='DEF ' then begin
	  try
		comp := TKiComponent.LoadFromFile(f, st);
		comp.filename := fn;
		KiComps.Add(comp)
	  except
		errsum := 1
	  end
	end
  end;

 goterror:
  closefile(f);
  if IOResult<>0 then;
  if (err<>0) or (errsum<>0)
	then application.MessageBox('Errors happened while reading.','KiLib Manager');

  // if exists a DCM file, try to load it
  fn := ChangeFileExt(fn, '.DCM');
  if FileExists(fn) then LoadDCM(KiComps, fn);

  refresh;
  fmMain.tmRefwin.Enabled := true
end;

constructor TfmLibrary.create(AOwner: TComponent);
begin
  inherited;
  KiComps := TList.create;
  DoLayout2;
end;

procedure TfmLibrary.LoadAskFile;
begin
  if not dlgOpen.Execute then begin
	close;
	exit
  end;
  LoadFile(dlgOpen.FileName);
end;

function CmpPassFilter(cmp: TKiComponent; flt: string): boolean;
begin
  result := true;
  if flt='' then exit;

  // no asterisks: simple filter
  if pos('*', flt)=0 then begin
	if (pos(flt, uppercase(cmp.name))=0) and (pos(flt, uppercase(cmp.aliases))=0) and (pos(flt, uppercase(cmp.keywords))=0) then result := false;
	exit
  end;

  result := StrLike(uppercase(cmp.name), flt) or pattern(uppercase(cmp.aliases), flt) or pattern(uppercase(cmp.keywords), flt)
end;

procedure TfmLibrary.Refresh;
var
  i		: integer;
  cmp	: TKiComponent;
  flt	: string;
  st	: string;

begin
  sgNames.Cells[0,0] := 'Name';
  sgNames.Cells[1,0] := 'Aliases';
  sgNames.Cells[2,0] := 'Keywords';
  sgNames.Cells[3,0] := 'index';
  sgNames.ColWidths[3] := 0;
  sgnames.RowCount := 2;
  flt := uppercase(edFilter.Text);
  for i := 0 to KiComps.Count-1 do begin
	cmp := TKiComponent(KiComps[i]);
	if not CmppassFilter(cmp, flt) then continue;
	sgNames.Cells[0, sgnames.RowCount-1] := cmp.name;
	sgNames.Cells[1, sgnames.RowCount-1] := cmp.aliases;
	sgNames.Cells[2, sgnames.RowCount-1] := cmp.Keywords;
	sgNames.Cells[3, sgnames.RowCount-1] := inttostr(i);
	sgNames.RowCount := sgNames.RowCount+1
  end;
  sgNames.Cells[0, sgnames.RowCount-1] := '';
  sgNames.Cells[1, sgnames.RowCount-1] := '';
  sgNames.Cells[2, sgnames.RowCount-1] := '';

  if sgnames.Rowcount>2 then sgnames.RowCount := sgnames.RowCount-1;
  SetModiCaption;

  sgNamesClick(NIL);
  fmMain.tmRefwin.Enabled := true
end;

procedure TfmLibrary.SetModiCaption;
var
  st	: string;

begin
  st := '';
  if libmodified then st := '* ';
  if libname=''
	then st := st + ' (noname): '
	else st := st + libpath+'\'+libname+': ';
  st := st+inttostr(sgnames.Rowcount-1)+'/'+inttostr(KiComps.count)+' components.';
  if edFilter.text<>''
	then st := st+' (filtered)';
  caption := st
end;

procedure TfmLibrary.FormCloseQuery(Sender: TObject; var CanClose: Boolean);
var
  res	: integer;

begin
  if libmodified then begin
	if WindowState=wsMinimized then WindowState := wsNormal;
	BringToFront;
	res := application.MessageBox('Data modified. Want to save first?', 'Closing library', MB_YESNOCANCEL);
	if res=idCANCEL then begin
	  canclose := false;
	  exit
	end;
	if res=idNO then exit;

	save;
	if libmodified then canclose := false
  end
end;

procedure TfmLibrary.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  if (winkcdel=self) or (winkcput=self) then begin
	winkcdel := NIL;
	winkcput := NIL;
  end;
  fmMain.tmRefwin.Enabled := true;
  Action := caFree
end;

procedure TfmLibrary.edFilterChange(Sender: TObject);
begin
  tmRefresh.Enabled := true
end;

procedure TfmLibrary.tmRefreshTimer(Sender: TObject);
begin
  tmRefresh.Enabled := false;
  refresh
end;

procedure TfmLibrary.sgNamesClick(Sender: TObject);
var
  i		: integer;

begin
  ActComp := NIL;
  intmodifying := true;
  panx := 0;
  pany := 0;
  ckDeMorgan.Enabled := false;
  ckDeMorgan.Checked := false;

  if sgNames.Cells[0, sgNames.Row]='' then begin
	lPinCount.Caption := '-';
	lUnits.Caption := '-';
	cbUnits.Text := ' -- ';
	cbUnits.Enabled := false;
	edAliases.Text := '';
	edKeywords.Text := '';
	edDescription.Text := '';
	lFilename.Caption := '';
  end else begin
	ActComp := KiComps[strtoint(sgNames.cells[3,sgNames.Row])];
	lFilename.Caption := '('+ActComp.filename+')';
	lPinCount.Caption := inttostr(ActComp.pins.Count);
	i := ActComp.unitcount;
	if i<1 then i := 1;
	lUnits.Caption := inttostr(i);
	if i=1 then begin
	  cbUnits.Enabled := false;
	  cbUnits.Clear;
	  cbUnits.items.Add(' -- ');
	  cbUnits.ItemIndex := 0;
	end else begin
	  cbUnits.Clear;
	  for i := 1 to ActComp.unitcount do cbUnits.Items.Add('Unit '+chr(64+i));
	  cbUnits.ItemIndex := 0;
	  cbUnits.Enabled := true;
	end;
	ckNames.Checked := ActComp.drawpinname;
	ckNumbers.Checked := ActComp.drawpinnum;
	edAliases.Text := ActComp.aliases;
	edKeywords.Text := ActComp.keywords;
	edDescription.Text := ActComp.description;
  end;
  pbComp.Invalidate;
  intmodifying := false;
end;

function creategoodfont(dimension, angle: integer; fontface: string): THandle;
begin
  result := CreateFont(
	-dimension,		// logical height of font
	0,				// logical average character width
	angle,			// angle of escapement
	0,				// base-line orientation angle
	300,			// font weight: slightly light
	0,				// italic attribute flag
	0,				// underline attribute flag
	0,				// strikeout attribute flag
	DEFAULT_CHARSET,		// character set identifier
	OUT_DEFAULT_PRECIS,		// output precision
	CLIP_DEFAULT_PRECIS,	// clipping precision
	PROOF_QUALITY	,		// output quality
	FF_SWISS,				// pitch and family
	PChar(fontface)			// pointer to typeface name string
	);
end;

procedure TfmLibrary.pbCompPaint(Sender: TObject);
var
  i,k	: integer;
  st,lav: string;
  cv	: TCanvas;
  r		: TRect;
  ts	: integer;

  function scl(fnum: integer): integer;
  begin
	result := tbZoom.position*strtointdef(fieldnum(fnum, st),0) div 100;
  end;

  function scx(fnum: integer): integer;
  begin
	result := scl(fnum) + pbComp.Width div 2 + panx;
  end;

  function scy(fnum: integer): integer;
  begin
	result := pbComp.height div 2 - scl(fnum) + pany;
  end;

var
  pinfont		: THandle;
  pinlen		: integer;
  theight		: integer;
  orient		: integer;
  tx,ty			: integer;
  points		: array[1..512] of TPoint;
  morgok		: boolean;

begin
  cv := pbComp.Canvas;
  cv.brush.Color := clWhite;
  cv.Pen.Color := clRed;
  cv.FillRect(rect(0,0, pbComp.Width, pbComp.Height));
  if ActComp=NIL then exit;
  cv.Font.Name := 'Arial';
  cv.Font.Color := clRed;
  morgok := false;

  // graphics
  for i := 0 to ActComp.graphics.Count-1 do begin
	st := PAnsiString(ActComp.graphics[i])^;
	case st[1] of
	  'S': begin
		// rectangle: S startx starty endx endy unit convert ltrait  cc
		if cbUnits.Enabled then begin
		  // does this pin belong to the unit?
		  k := strtointdef(fieldnum(5, st),0);
		  if k<>0 then if cbUnits.ItemIndex+1 <> k then continue
		end;
		// check DeMorgan
		k := strtointdef(fieldnum(6, st),0);
		if k<>0 then begin
		  if k=2 then morgok := true;
		  if (k=2) xor ckDeMorgan.Checked then continue;
		end;

		r.Left := scx(1);
		r.Top  := scy(2);
		r.Right := scx(3);
		r.Bottom := scy(4);
		cv.Brush.Style := bsClear;
		cv.Rectangle(r)
		end;
	  'A': begin
		// arc: posx posy radius start end part convert ltrait start_pointX start_pointY end_pointX end_pointY  cc
		//      0    0    75     900   0   0    1       0      N 0 75 75 0
		if cbUnits.Enabled then begin
		  // does this pin belong to the unit?
		  k := strtointdef(fieldnum(6, st),0);
		  if k<>0 then if cbUnits.ItemIndex+1 <> k then begin
			r.Left := 0;
			continue
		  end
		end;
		// check DeMorgan
		k := strtointdef(fieldnum(7, st),0);
		if k<>0 then begin
		  if k=2 then morgok := true;
		  if (k=2) xor ckDeMorgan.Checked then continue;
		end;
		r.Left := scx(1);
		r.Top  := scy(2);
		tx := scl(3);	// radius

		// se l'angolo DEcresce, fare questo:
		if strtointdef(fieldnum(4, st),0)>strtointdef(fieldnum(5, st),0) then begin
		  points[1].X := scx(10);	// startx
		  points[1].Y := scy(11);	// startY
		  points[2].X := scx(12);	// endx
		  points[2].Y := scy(13);	// endY
		end else begin
		  points[2].X := scx(10);	// startx
		  points[2].Y := scy(11);	// startY
		  points[1].X := scx(12);	// endx
		  points[1].Y := scy(13);	// endY
		end;
		cv.Brush.Style := bsClear;
		cv.Arc(r.Left-tx, r.Top-tx, r.left+tx, r.Top+tx, points[2].x, points[2].y, points[1].x, points[1].y)
		end;

	  'C': begin
		// circle: C posx posy radius unit convert ltrait cc
		if cbUnits.Enabled then begin
		  // does this pin belong to the unit?
		  k := strtointdef(fieldnum(4, st),0);
		  if k<>0 then if cbUnits.ItemIndex+1 <> k then continue
		end;
		// check DeMorgan
		k := strtointdef(fieldnum(5, st),0);
		if k<>0 then begin
		  if k=2 then morgok := true;
		  if (k=2) xor ckDeMorgan.Checked then continue;
		end;
		r.Left := scx(1);
		r.Top  := scy(2);
		tx := scl(3);
		cv.Brush.Style := bsClear;
		cv.Arc(r.Left-tx, r.Top-tx, r.left+tx, r.Top+tx, 0, 0, 0, 0)
		end;
	  'P': begin
		// polygon: P Nb parts convert ltrait x0 y0 x1 y1 xi yi cc
		if cbUnits.Enabled then begin
		  // does this pin belong to the unit?
		  k := strtointdef(fieldnum(2, st),0);
		  if k<>0 then if cbUnits.ItemIndex+1 <> k then continue
		end;
		// check DeMorgan
		k := strtointdef(fieldnum(3, st),0);
		if k<>0 then begin
		  if k=2 then morgok := true;
		  if (k=2) xor ckDeMorgan.Checked then continue;
		end;
		tx := strtointdef(fieldnum(1, st),0);	// number of points
		k := 5;
		for ty := 1 to tx do begin
		  points[ty].X := scx(k);
		  inc(k);
		  points[ty].Y := scy(k);
		  inc(k)
		end;
		case fieldnum(k,st)[1] of
		  'N', ' ': begin
			// inc(tx);
			// points[tx] := points[1];
			cv.Polyline(slice(points, tx));
			end;
		  'f': begin
			cv.Brush.Color := clYellow;
			cv.Polygon(slice(points, tx));
			end;
		  'F': begin
			cv.Brush.Color := cv.Pen.Color;
			cv.Polygon(slice(points, tx));
			end;
		  end;
		end;
	  'T': begin
		// text: T orientation posx posy dimension unit convert Text
		if cbUnits.Enabled then begin
		  // does this pin belong to the unit?
		  k := strtointdef(fieldnum(5, st),0);
		  if k<>0 then if cbUnits.ItemIndex+1 <> k then continue
		end;
		// check DeMorgan
		k := strtointdef(fieldnum(6, st),0);
		if k<>0 then begin
		  if k=2 then morgok := true;
		  if (k=2) xor ckDeMorgan.Checked then continue;
		end;
		k := strtointdef(fieldnum(1, st),0);
		if k=1 then k := 900;
		tx := scl(4);
		pinfont := creategoodfont(tx, k, 'Arial');
		SelectObject(cv.Handle, pinfont);
		comando := st;
		separa;
		separa;
		separa;
		separa;
		separa;
		separa;
		separa;
		separa;
		TextOut(
		  cv.Handle,		// handle of device context
		  scx(2),			// x-coordinate of starting position
		  scy(3)-ty,		// y-coordinate of starting position
		  PChar(Comando),	// address of string
		  length(Comando)	// number of characters in string
		);
		DeleteObject(pinfont)
		end;
	end
  end;

  // plot pins
  ts := tbZoom.position*16 div 100;		// for circles
  // cv.Font.Height := scl(6)*7 div 5;	// number text size;
  cv.Brush.Style := bsClear;			// for textout()

  for i := 0 to ActComp.pins.Count-1 do begin
	// BALANCE 5 500 -100 300 L 40 40 1 1 P
	// a pin: name number posx posy length orientation Snum Snom unit convert Etype [shape]
	// the first "X"^^ is deleted
	st := PAnsiString(ActComp.pins[i])^;
	if cbUnits.Enabled then begin
	  // does this pin belong to the unit?
	  k := strtointdef(fieldnum(8, st),0);
	  if k<>0 then if cbUnits.ItemIndex+1 <> k then continue;
	end;
	// check DeMorgan
	k := strtointdef(fieldnum(9, st),0);
	if k<>0 then begin
	  if k=2 then morgok := true;
	  if (k=2) xor ckDeMorgan.Checked then continue;
	end;

	cv.Font.Color := clRed;
	// fmMain.sbbStatus.Panels[4].Text := st;
	r.Left := scx(2);
	r.top  := scy(3);
	// little circle
	cv.Arc(r.Left-ts, r.Top-ts, r.Left+ts, r.Top+ts, 0,0,0,0);
	orient := 0;
	theight := scl(6)*9 div 6;
	if theight=0 then theight := 1;
	tx := r.Left;
	ty := r.top-theight-2;
	pinlen := scl(4);
	case upcase(fieldnum(5, st)[1]) of
	  'R': begin r.right := r.Left+pinlen+1; r.Bottom := r.Top; tx := (r.Left+r.Right) div 2; end;
	  'L': begin r.Right := r.Left-pinlen-1; r.Bottom := r.Top; tx := (r.Left+r.Right) div 2+theight; end;
	  'D': begin r.bottom:= r.top+pinlen+1;  r.right := r.left; orient := 900; ty := (r.Top+r.Bottom) div 2+theight; end;
	  'U': begin r.bottom:= r.top-pinlen-1;  r.right := r.left; orient := 900; ty := (r.Top+r.Bottom) div 2+theight ; end;
	end;
	// pin lines
	cv.moveto(r.Left, r.Top);
	cv.lineto(r.Right, r.Bottom);

	// pin number
	if ckNumbers.Checked then begin
	  pinfont := creategoodfont(theight, orient, cv.Font.Name);
	  SelectObject(cv.Handle, pinfont);
	  st := fieldnum(1,st);
	  TextOut(
		cv.Handle,	// handle of device context
		tx-theight-2,	// x-coordinate of starting position
		ty,			// y-coordinate of starting position
		PChar(st),	// address of string
		length(st)	// number of characters in string
	  );
	  DeleteObject(pinfont);
	  st := PAnsiString(ActComp.pins[i])^;	// get the string again...
	end;

	// pin name
	if (fieldnum(0,st)<>'~') and ckNames.Checked then begin
	  cv.Font.Color := clRed;
	  k := scl(7)*4 div 3;
	  if k=0 then k := 1;
	  pinfont := creategoodfont(k, orient, 'Arial');
	  cv.Font.Height := -k;
	  k := cv.TextWidth(fieldnum(0,st));
	  SelectObject(cv.Handle, pinfont);
	  case upcase(fieldnum(5, st)[1]) of
		'R': begin inc(tx,pinlen div 2+4) end;
		'L': begin dec(tx,pinlen div 2+k+14) end;
		'D': begin dec(tx,scl(7) div 2+2); ty := r.bottom+k-4 end;
		'U': begin dec(tx,scl(7) div 2+2); ty := r.Bottom-scl(7)+2 end;
	  end;
	  st := fieldnum(0,st);
	  TextOut(
		cv.Handle,	// handle of device context
		tx,			// x-coordinate of starting position
		ty+theight div 2,			// y-coordinate of starting position
		PChar(st),	// address of string
		length(st)	// number of characters in string
	  );
	  DeleteObject(pinfont)
	end
  end;

  // fields
  cv.Font.Color := clBlue;
  for i := 0 to ActComp.fields.Count-1 do begin
	// Fn text posx posy dimension orientation visibility hjustify vjustify/italic/bold name
	// n=0: reference; n=1: value; other... special
	st := PAnsiString(ActComp.fields[i])^;
	lav := copy(fieldnum(0, st),2,2);
	if not (strtointdef(lav,0) in [0..2]) then continue;
	r.Left := scx(2);
	r.Top  := scy(3);
	theight := scl(4)*8 div 6;
	if theight=0 then theight := 1;
	k := 0;
	if fieldnum(5, st)='V' then k := 900;
	cv.Font.Height := -theight;
	lav := fieldnum(1, st);
	lav := copy(lav, 2, length(lav)-2);
	tx := cv.TextWidth(lav);
	if k=0 then begin
	  // horizontal
	  case fieldnum(7, st)[1] of
		'R': dec(r.Left, tx);
		'C': dec(r.Left, tx div 2)
	  end;
	  case fieldnum(8, st)[1] of
		'T': dec(r.Top, theight);
		'C': dec(r.Top, theight div 2)
	  end;
	end else begin
	  // vertical
	  case fieldnum(7, st)[1] of
		'R': inc(r.top, tx);
		'C': inc(r.top, tx div 2)
	  end;
	  case fieldnum(8, st)[1] of
		'T': dec(r.left, theight);
		'C': dec(r.left, theight div 2)
	  end;
	end;

	pinfont := creategoodfont(theight, k, 'Arial');
	SelectObject(cv.Handle, pinfont);
	TextOut(
	  cv.Handle,		// handle of device context
	  r.left,			// x-coordinate of starting position
	  r.top,			// y-coordinate of starting position
	  PChar(lav),		// address of string
	  length(lav)		// number of characters in string
	  );
	DeleteObject(pinfont)
  end;

  // if the part has no deMorgan representation, delete it
  if morgok then begin
	ckDeMorgan.Enabled := true;
  end else begin
	(*
	ckDeMorgan.Enabled := false;
	if ckDeMorgan.Checked then begin
	  ckDeMorgan.checked := false;
	  pbComp.Invalidate
	end
	*)
  end
end;

procedure TfmLibrary.tbZoomChange(Sender: TObject);
begin
  lZoom.Caption := inttostr(tbZoom.Position*48 div 70*10)+'%';
  pbComp.Invalidate
end;

procedure TfmLibrary.DoLayout1;
begin
  pnList.Align := alTop;
  Splitter1.Align := alTop;
  pnComp.Align := alClient;

  pnView.Align := alLeft;
  pnView.Width := 188;
  pnZoom.Align := alTop;
  pnZoom.Height := 78;
  pnMorgan.Top := pnZoom.Height;

  width := width-1;
  width := width+1;
end;

procedure TfmLibrary.DoLayout2;
begin
  pnList.Align := alLeft;
  pnList.Width := 300;
  Splitter1.Align := alLeft;
  Splitter1.Left := pnList.Width+20;
  pnComp.Align := alClient;

  pnView.Align := alTop;
  pnView.Height := 78;
  pnZoom.Align := alLeft;
  pnZoom.Width := 186
end;

procedure TfmLibrary.sgNamesDragOver(Sender, Source: TObject; X, Y: Integer; State: TDragState; var Accept: Boolean);
begin
  accept := false;
  if TControl(source).Owner=self then exit;
  Accept := TControl(sender).Owner<>TControl(source).Owner
end;

procedure TfmLibrary.sgNamesDragDrop(Sender, Source: TObject; X, Y: Integer);
var
  src	: TfmLibrary;
  op	: integer;		// 0=move, 1=copy
  i		: integer;
  acomp	: TKiComponent;
  st,sb	: integer;

begin
  if source=NIL then exit;
  src := TfmLibrary(TControl(Source).owner);
  // "original" operation
  op := 1;
  if TControl(source).name='lCopy' then op := 1;
  if TControl(source).name='lMove' then op := 0;
  if sender<>NIL then if TControl(sender).name='lMove' then op := 0;
  if sender<>NIL then if TControl(sender).name='lCopy' then op := 1;
  st := src.sgNames.Selection.top;
  sb := src.sgNames.Selection.Bottom;

  kcBoard.Clear;
  winkcput := self;
  winkcdel := NIL;

  self.BringToFront;
  if op=0 then begin
	if Application.MessageBox(PChar('Sure to move here '+inttostr(sb-st+1)+' items?'), 'Drag-Move', MB_YESNO+MB_DEFBUTTON2)<>idYES then exit;
	fmMain.mess('Moving '+inttostr(sb-st+1)+' items...');
	winkcdel := src
  end else begin
	if sb-st>20 then begin
	  if Application.MessageBox(PChar('Sure to copy here '+inttostr(sb-st+1)+' items?'), 'Drag-Move', MB_YESNO+MB_DEFBUTTON2)<>idYES then exit;
	end;
	fmMain.mess('Copying '+inttostr(sb-st+1)+' items...');
  end;

  libmodified := true;

  // copy items
  for i := st to sb do begin
	with src do acomp :=  KiComps[strtoint(sgNames.cells[3,i])];
	KiComps.Add(acomp);
	kcBoard.Add(acomp)
  end;
  refresh;

  if op=0 then with src do begin
	// delete from old
	for i := sb downto st do KiComps.Delete(strtoint(sgNames.cells[3,i]));
	libmodified := true;
	refresh;
	fmMain.mnUndo.Caption := '&Undo: move'
  end else fmMain.mnUndo.Caption := '&Undo: duplicate';
  fmMain.mnUndo.Enabled := true;
end;

procedure TfmLibrary.Save;
var
  f		: textfile;
  i		: integer;
  cmp	: TKiComponent;

begin
  if libname='' then begin
	if not dlgSave.Execute then exit;
	libname := ExtractFileName(dlgSave.FileName);
	libpath := ExtractFileDir(dlgSave.FileName);
	refresh;
  end;
  fmMain.mess('Saving '+libname+' ...');
  assignfile(f, libpath+'\'+libname);
  rewrite(f);

  // fix intestation
  writeln(f, 'EESchema-LIBRARY Version 2.3');
  for i := 0 to KiComps.Count-1 do begin
	cmp := KiComps[i];
	cmp.WriteOut(f);
  end;
  closefile(f);
  //////////

  // save DCM
  assignfile(f, ChangeFileExt(libpath+'\'+libname, '.dcm'));
  rewrite(f);

  writeln(f, 'EESchema-DOCLIB  Version 2.0');
  for i := 0 to KiComps.Count-1 do begin
	writeln(f, '#');
	cmp := KiComps[i];
	writeln(f, '$CMP ', cmp.name);
	if cmp.description<>'' then writeln(f, 'D ', cmp.description);
	if cmp.keywords<>'' then writeln(f, 'K ', cmp.keywords);
	if cmp.documents<>'' then writeln(f, 'F ', cmp.documents);
	writeln(f, '$ENDCMP')
  end;
  writeln(f, '#');
  writeln(f, '#End Doc Library');
  closefile(f);

  fmMain.mess('Saved '+libname);
  libmodified := false;
  refresh
end;

procedure TfmLibrary.SaveAs;
begin
  if not dlgSave.Execute then exit;
  libname := ExtractFileName(dlgSave.FileName);
  libpath := ExtractFileDir(dlgSave.FileName);
  save
end;

procedure TfmLibrary.mnDelete1Click(Sender: TObject);
var
  i		: integer;
  cmp	: TKiComponent;

begin
  if sgNames.cells[0,1]='' then exit;

  kcBoard.Clear;
  winkcput := NIL;
  winkcdel := self;

  fmMain.mess('Deleting '+inttostr(sgNames.Selection.Bottom-sgNames.Selection.Top+1)+' items...');
  for i := sgNames.Selection.Bottom downto sgNames.Selection.top do begin
	cmp :=  KiComps[strtoint(sgNames.cells[3,i])];
	kcBoard.Add(cmp);
	KiComps.Remove(cmp)
  end;
  fmMain.mnUndo.Enabled := true;
  fmMain.mnUndo.Caption := '&Undo: delete';
  libmodified := true;

  refresh
end;

procedure TfmLibrary.mnCut1Click(Sender: TObject);
var
  i		: integer;
  cmp	: TKiComponent;

begin
  if sgNames.cells[0,1]='' then exit;

  kcBoard.Clear;
  winkcput := NIL;
  winkcdel := self;

  fmMain.mess('Cutting '+inttostr(sgNames.Selection.Bottom-sgNames.Selection.Top+1)+' items...');
  for i := sgNames.Selection.Bottom downto sgNames.Selection.top do begin
	cmp :=  KiComps[strtoint(sgNames.cells[3,i])];
	kcBoard.Add(cmp);
	KiComps.Remove(cmp)
  end;
  fmMain.mnUndo.Enabled := true;
  fmMain.mnUndo.Caption := '&Undo: cut';
  libmodified := true;

  refresh
end;

procedure TfmLibrary.mnCopy1Click(Sender: TObject);
var
  i		: integer;
  cmp	: TKiComponent;

begin
  if sgNames.cells[0,1]='' then exit;

  kcBoard.Clear;
  winkcput := NIL;
  winkcdel := NIL;

  fmMain.mess('Copying '+inttostr(sgNames.Selection.Bottom-sgNames.Selection.Top+1)+' items...');
  for i := sgNames.Selection.Bottom downto sgNames.Selection.top do begin
	cmp :=  KiComps[strtoint(sgNames.cells[3,i])];
	kcBoard.Add(cmp);
  end;
  fmMain.mnUndo.Enabled := false;
  refresh
end;

procedure TfmLibrary.pnListResize(Sender: TObject);
begin
  with sgnames do begin
	colwidths[1] := (sgnames.ClientWidth-colwidths[0]) div 2-3;
	colwidths[2] := colwidths[1]
  end
end;

procedure TfmLibrary.mnPaste1Click(Sender: TObject);
var
  i		: integer;

begin
  if kcBoard.Count=0 then exit;

  for i := 0 to kcBoard.Count-1 do KiComps.Add(kcBoard[i]);

  winkcput := self;
  winkcdel := NIL;
  fmMain.mnUndo.Enabled := true;
  fmMain.mnUndo.Caption := '&Undo: paste';
  libmodified := true;
  refresh
end;

procedure TfmLibrary.mnSortbyname1Click(Sender: TObject);
var
  cmp	: TKiComponent;
  i		: integer;
  swp	: boolean;

begin
  repeat
	swp := true;
	for i := 0 to KiComps.Count-2 do begin
	  cmp := KiComps[i];
	  if cmp.name>TKiComponent(KiComps[i+1]).name then begin
		KiComps[i] := KiComps[i+1];
		KiComps[i+1] := cmp;
		swp := false;
		break
	  end
	end
  until swp;
  libmodified := true;
  refresh
end;

procedure TfmLibrary.edFilterKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if key=VK_DOWN then sgNames.SetFocus;
  if key=VK_ESCAPE  then edFilter.SelectAll
end;

procedure TfmLibrary.sgNamesKeyDown(Sender: TObject; var Key: Word; Shift: TShiftState);
begin
  if key=VK_ESCAPE then begin
	edFilter.SetFocus;
	edFilter.SelectAll
  end
end;

procedure TfmLibrary.edKeywordsChange(Sender: TObject);
begin
  if not intmodifying then begin
	libmodified := true;
	ActComp.aliases := edAliases.Text;
	ActComp.description := edDescription.Text;
	ActComp.keywords    := edKeywords.Text;
	SetModiCaption
  end
end;

procedure TfmLibrary.pbCompMouseDown(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  pandown := true;
  tbZoom.SetFocus;
  opx := x;
  opy := y;
  if Button=mbLeft then begin
	inc(panx,pbComp.Width div 2-x);
	inc(pany,pbComp.Height div 2-y);
  end else begin
	panx := 0;
	pany := 0;
  end;
  pbComp.Invalidate
end;

procedure TfmLibrary.pbCompMouseUp(Sender: TObject; Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  pandown := false;
end;

procedure TfmLibrary.pbCompMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
begin
  if (x<0) or (y<0) then pandown := false;
  if (x>pbComp.ClientWidth) or (y>pbComp.ClientHeight) then pandown := false;
  if not pandown then exit;

  inc(panx, x-opx);
  inc(pany, y-opy);

  opx := x;
  opy := y;

  pbComp.Invalidate
end;

end.

