unit txtintface;

interface
Uses Classes;

type
  // DEF name reference unused text_offset draw_pinnumber draw_pinname unit_count units_locked
  TKiComponent	= class(TObject)
	name		: string;		// component name in library (74LS02 ...)
	filename	: string;		// only used in search function
	reference	: string;		// Reference ( U, R, IC .., which become U3, U8, R1, R45, IC4...)
	aliases		: string;		// list of names
	description	: string;		// from .dcm
	keywords	: string;		// from .dcm
	documents	: string;		// from .dcm
	textoffset	: integer;      // offset for pin name position
	drawpinnum	: boolean;		// Y (display pin number) ou N (do not display pin number).
	drawpinname	: boolean;		// Y (display pin name) ou N (do not display pin name).
	unitcount	: integer;		// Number of part (or section) in a component package.
	unitlocked	: char;			// L=not identical-cannot be swapped) or F=identical, can be swapped only if unit_count > 1
	optionflag	: char;			// N (normal) or P (component type "power")
	fields		: TList;		// list of fields
	graphics	: TList;
	pins		: TList;		// list of pins / graphic elements

	constructor	create;
	constructor	LoadFromFile(var f: TextFile; st: string);
	procedure	WriteOut(var f: textFile);

	protected
	  procedure	wrout(var f: TextFile; st: string); overload;
	  procedure	wrout(var f: TextFile; nn: integer); overload;
	  procedure	wrout(var f: TextFile; bb: boolean); overload;
	end;

var
  comando	: string;		// for function separa

Function  Separa: string;	// extract space-seaparated fields from "comando"

function  fieldnum(num: integer; st: string): string;
	// gets Nth field from "st"

function  IsKiCadLibProlog(st: string): boolean;
	// returns whether the string is the correct prolog

function  IsKiCadDcmProlog(st: string): boolean;
	// returns whether the string is the correct prolog

function  StrLike(st, reg: string): boolean;
	// checks if "st" matches "reg" using "*" and "?"
	// using either of the following

function  Pattern(st, reg: string): boolean;
	// checks if "st" matches "pattern" using "*" (my routine)

function  like(const S, Substr: AnsiString): Boolean;
	// checks if "st" matches "pattern" using "*" and "?" (someone else's routine)

function SearchPassFilter(field, flt: string): boolean;
	// search for "flt" in "field"

procedure LoadDCM(KiComps: TList; fn: string);

implementation
Uses sysutils;

function IsKiCadLibProlog(st: string): boolean;
begin
  result := pos('EESchema-LIBRARY ', st)<>0
end;

function IsKiCadDcmProlog(st: string): boolean;
begin
  result := pos('EESchema-DOCLIB ', st)<>0
end;

function  StrLike(st, reg: string): boolean;
begin
  result := pattern(st, reg)
end;

procedure	TKiComponent.WriteOut(var f: textFile);
var
  i		: integer;

begin
  writeln(f, '#');
  writeln(f, '# '+name);
  writeln(f, '#');
  // DEF name reference unused text_offset draw_pinnumber draw_pinname unit_count units_locked
  // DEF 1408 U 0 10 Y Y 1 F N
  // DEF PIC16F84 U*** 0 40 Y Y 1 0 N
  write(f, 'DEF ', name, ' ', reference);
  wrout(f, '0');	// unused
  wrout(f, textoffset);
  wrout(f, drawpinnum);		wrout(f, drawpinname);
  wrout(f, unitcount);		wrout(f, unitlocked);
  wrout(f, optionflag);
  writeln(f);

  // fields
  for i := 0 to fields.Count-1 do writeln(f, PAnsiString(fields[i])^);

  // alias
  writeln(f, 'ALIAS ', aliases);

  // draw
  writeln(f, 'DRAW ');
	// graphics
	for i := 0 to graphics.Count-1 do writeln(f, PAnsiString(graphics[i])^);
	// pins
	for i := 0 to pins.Count-1 do writeln(f, 'X ', PAnsiString(pins[i])^);
  writeln(f, 'ENDDRAW');
  writeln(f, 'ENDDEF');
end;

procedure TKiComponent.wrout(var f: TextFile; st: string);
begin
  write(f, ' ', st)
end;

procedure TKiComponent.wrout(var f: TextFile; nn: integer);
begin
  write(f, ' ', nn)
end;

procedure TKiComponent.wrout(var f: TextFile; bb: boolean);
begin
  if bb
	then write(f, ' Y')
	else write(f, ' N')
end;

Function Separa: string;
var
  b : integer;

begin
  while copy(Comando,1,1)=' ' do delete(Comando,1,1);
  if copy(comando,1,1)='"' then begin
	b := pos('"',copy(comando,2,8192));
	if boolean(b) then inc(b,2)
  end else b := pos(' ',Comando);
  if b=0 then b := length(Comando)+1;
  Separa := copy(Comando,1,b-1);
  Delete(Comando,1,b)
end;

// returns the n-th field of the string
function fieldnum(num: integer; st: string): string;
var
  sav	: string;

begin
  sav := comando;
  comando := st;
  while (num>-1) do begin
	result := separa;
	dec(num)
  end;
  if result='' then result := ' ';
  comando := sav
end;

constructor TKiComponent.create;
begin
  fields   := TList.Create;
  pins     := TList.Create;
  graphics := TList.Create;
end;


constructor TKiComponent.LoadFromFile(var f: TextFile; st: string);
var
  err			: integer;
  ingraphics	: boolean;

  afield	: PAnsiString;
  first		: string;


  function getline: boolean;	// true if error
  begin
	result := true;
	if eof(f) then begin
	  err := 1;
	  exit
	end;
	readln(f, st);
	err := IOResult;
	if err=0 then result := false
  end;

  // every line in DRAW-ENDDRAW comes here
  procedure readgraphic;
  begin
	if first='ENDDRAW' then begin
	  ingraphics := false;
	  exit
	end;
	// is it a pin?
	new(afield);
	if first[1]='X' then begin
	  afield^ := comando;	// only get data, not "X ..."
	  pins.Add(afield)
	end else begin
	  afield^ := st;
	  graphics.Add(afield)
	end
  end;

label goterror;

begin
  fields := TList.Create;
  pins   := TList.Create;
  graphics := TList.Create;

  err := 0;
  ingraphics := false;

  // DEF 1408 U 0 10 Y Y 1 F N
  // DEF PIC16F84 U*** 0 40 Y Y 1 0 N
  // DEF name reference unused text_offset draw_pinnumber draw_pinname unit_count units_locked option_flag
  comando := st;
  separa;					// DEF
  name := separa;			// PIC16F...
  reference	:= separa;
  separa;					// unused
  textoffset := StrToIntDef(separa, 0);
  drawpinnum := uppercase(separa)='Y';
  drawpinname:= uppercase(separa)='Y';
  unitcount  := strtointdef(separa, 1);
  if unitcount>1 then begin
	// they can be locked
	unitlocked := uppercase(separa)[1];
  end else begin
	separa;
	unitlocked := 'F';
  end;
  optionflag := separa[1];

  repeat
	if getline then goto goterror;
	comando := st;
	first := separa;
	if ingraphics then readgraphic else begin
	  if first='ALIAS' then aliases := comando;
	  if first='DRAW' then ingraphics := true;
	  if copy(first,1,1)='F' then begin
		// fields F0, F1...
		new(afield);
		afield^ := st;
		fields.Add(afield)
	  end
	end
  until st='ENDDEF';
  exit;

 goterror:
  fail
end;

procedure LoadDCM(KiComps: TList; fn: string);
var
  lin	: string;
  fld	: string;
  cmp	: TKiComponent;
  f		: TextFile;
  i		: integer;

begin
  assign(f, fn);
  reset(f);
  if IOResult<>0 then exit;

  cmp := NIL;

  while not eof(f) do begin
	if IOresult<>0 then break;
	readln(f, lin);
	comando := lin;
	fld := separa;
	if cmp<>NIL then begin
	  // put in fields
	  if fld='D' then cmp.description := comando
	  else if fld='K' then cmp.keywords := cmp.keywords + comando
	  else if fld='F' then cmp.documents := cmp.documents + comando
	  else if fld='#' then
	  else cmp := NIL
	end else begin
	  if fld='$CMP' then begin
		for i := 0 to KiComps.Count-1 do begin
		  if TKiComponent(KiComps[i]).name=comando then begin
			cmp := TKiComponent(KiComps[i]);
			break
		  end
		end
	  end
	end
  end;
  close(f);
  if IOResult<>0 then exit;
end;

function pattern(st,reg: string): boolean;
var
  i,k       : integer;

label
  ok, Asterisco;

label fine;

begin
  i := 1;
  k := 1;
  pattern := false;

  while (i<=length(st)) and (k<=length(reg)) do begin
	if reg[k]='*' then begin
	  // go to next
	  inc(k);
	  if k>length(reg) then goto ok;    // final asterisk - ok
	  while (i<=length(st)) and (st[i]<>reg[k]) do inc(i);
	  if i>length(st) then goto fine
	end else if (reg[k]='?') or (reg[k]=st[i]) then begin
	  inc(i);
	  inc(k)
	end else begin

  Asterisco:
	  // if there was an asterisk BEFORE, then returns 1; otherwise fail
	  if pos('*',copy(reg,1,k))=0 then goto fine;
	  if k>length(reg) then goto fine;
	  while reg[k]<>'*' do dec(k)
	end
  end;
  if i<=length(st) then goto asterisco;
  if k<=length(reg) then if (k<length(reg)) or (reg[k]<>'*') then goto fine;

ok:
  pattern := true;

fine:
end;

// more complete routine, not tested (and not used)
function like(const S, Substr: AnsiString): Boolean;
var
  StringPtr		: PAnsiChar;
  PatternPtr	: PAnsiChar;
  StringRes		: PAnsiChar;
  PatternRes	: PAnsiChar;
  index			: integer;

begin
  result := true;
  if SubStr = ''  then exit;
  if SubStr = '*' then exit;

  if S = '' then begin
	result := false;
	Exit
  end;
  index := 1;

  StringPtr := PAnsiChar(@S[Index]);
  PatternPtr := PAnsiChar(SubStr);
  StringRes := nil;
  PatternRes := nil;

  repeat
	repeat
      case PatternPtr^ of
        #0:
          begin
            Result := StringPtr^ = #0;
            if Result or (StringRes = nil) or (PatternRes = nil) then
              Exit;

            StringPtr := StringRes;
            PatternPtr := PatternRes;
            Break;
          end;
        '*':
          begin
            Inc(PatternPtr);
            PatternRes := PatternPtr;
            Break;
          end;
        '?':
          begin
            if StringPtr^ = #0 then
              Exit;
            Inc(StringPtr);
			Inc(PatternPtr);
		  end;
		else
          begin
            if StringPtr^ = #0 then
			  Exit;
            if StringPtr^ <> PatternPtr^ then
            begin
              if (StringRes = nil) or (PatternRes = nil) then
                Exit;
              StringPtr := StringRes;
              PatternPtr := PatternRes;
              Break;
            end
            else
            begin
              Inc(StringPtr);
              Inc(PatternPtr);
            end;
          end;
      end;
    until False;

    repeat
      case PatternPtr^ of
        #0:
          begin
            Result := True;
            Exit;
		  end;
		'*':
          begin
			Inc(PatternPtr);
			PatternRes := PatternPtr;
          end;
        '?':
          begin
            if StringPtr^ = #0 then
              Exit;
            Inc(StringPtr);
            Inc(PatternPtr);
          end;
        else
          begin
            repeat
              if StringPtr^ = #0 then
                Exit;
              if StringPtr^ = PatternPtr^ then
                Break;
              Inc(StringPtr);
            until False;
            Inc(StringPtr);
            StringRes := StringPtr;
            Inc(PatternPtr);
            Break;
		  end;
	  end
	until False
  until False
end;

function SearchPassFilter(field, flt: string): boolean;
begin
  result := true;
  if flt='' then exit;
  if field='' then begin
	result := false;
	exit
  end;

  // no asterisks: simple filter
  if pos('*', flt)=0 then begin
	result := pos(flt, uppercase(field))<>0;
	exit
  end;

  result := StrLike(uppercase(field), flt)
end;

end.

